/*
 * dSS11 led handling daemon
 * Copyright (c) 2011,2012 aizo ag, Switzerland
 *
 * Author: Johannes Winkelmann <johannes.winkelmann@aizo.com>
 *         Michael Tross <michael.tross@aizo.com>
 */

#include <string>
#include <vector>
#include <fstream>
#include <iostream>
#include <cstdlib>
#include <pthread.h>

#include "commandhandler.h"
#include "led.h"
#include "sequence.h"
#include "ledgroup.h"
#include "report.h"

struct segment {
  uint32_t time;
  uint32_t red;
  uint32_t green;
  uint32_t blue;
};

/*
 Code definition:

 Blink code definition consist of a sequence of states, each
 containing a duration and the RGB states of the LEDs

 Definitions must be ended with a state of duration 0; for
 convenience, an END_STATE macro exists

 Note that definitions containing a single non-termination state will
 not loop, i.e. the LEDs will be set once, and then the thread will
 terminate; however one still has to chose a non-zero duration, for
 obvious reasons
*/

#define END_CODE { 0, 0, 0, 0 }

static segment code_ready[] = {
  { 5000, 0, 1, 0 },
  END_CODE
};

static segment code_startup[] = {
  { 5000, 1, 1, 0 },
  END_CODE
};

static segment code_failure[] = {
  { 5000, 1, 0, 0 },
  END_CODE
};

static segment code_system_update[] = {
  { 500, 1, 1, 0 },
  { 500, 0, 0, 1 },
  END_CODE
};

static segment code_rainbow[] = {
  {  250, 0, 1, 0 },
  {  250, 1, 1, 0 },
  {  250, 1, 0, 0 },
  {  250, 1, 0, 1 },
  {  250, 0, 0, 1 },
  {  250, 0, 1, 1 },
  END_CODE
};

CommandHandler::CommandHandler(LedFactory* ledFactory)
  : m_ledFactory(ledFactory),
    m_sequence(0),
    m_state(eSTATE_STARTUP),
    m_oldState(eSTATE_STARTUP),
    m_ds485Ok(true),
    m_info2Ok(true),
    m_info3Ok(true),
    m_netOk(true),
    m_wifiOk(true),
    m_internetOk(true),
    m_hardwareErr(false),
    m_softwareErr(false),
    m_err2Err(false),
    m_thread(0)
{
}

CommandHandler::~CommandHandler()
{
  if (m_sequence) {
    m_sequence->stop();
    delete m_sequence;
  }
}

void CommandHandler::handle(const ledd_actions command)
{
  report(LOG_INFO, "Received command: %d - %s\n", command,
         ledd_action_to_string(command));
 
  // states:
  //   states:
  //     startup
  //     ready
  //     failure
  //     system_update_started/system_update_completed
  //     rainbow/rainbow_completed
  //   info substates (grey flashing)
  //     ds485ok/ds485_err
  //     info2_ok/info2_err
  //     info3_ok/info3_err
  //   network substates (magenta flashing)
  //     net_ok/net_err
  //     wifi_ok/wifi_err
  //     internet_ok/internet_err
  //   error substates (red flashing)
  //     hw_ok/hw_err
  //     sw_ok/sw_err
  //     err2_ok/err2_err
 
  // states
  if (command == ledd_ready) {
    if ((m_state == eSTATE_UPDATE) || (m_state == eSTATE_RAINBOW)) {
      m_oldState = eSTATE_READY;
    } else {
      m_state = eSTATE_READY;
    }
  } else if (command == ledd_failure) {
    if ((m_state == eSTATE_UPDATE) || (m_state == eSTATE_RAINBOW)) {
      m_oldState = eSTATE_FAILURE;
    } else {
      m_state = eSTATE_FAILURE;
    }
  } else if ((command == ledd_system_update_started) && (m_state != eSTATE_UPDATE)) {
    m_oldState = m_state;
    m_state = eSTATE_UPDATE;
  } else if ((command == ledd_system_update_completed) && (m_state == eSTATE_UPDATE)) {
    m_state = m_oldState;
  } else if ((command == ledd_rainbow) && (m_state != eSTATE_RAINBOW)) {
    m_oldState = m_state;
    m_state = eSTATE_RAINBOW;
  } else if ((command == ledd_rainbow_completed) && (m_state == eSTATE_RAINBOW)) {
    m_state = m_oldState;
  }
 
  // info substates
  else if (command == ledd_ds485_ok) {
    m_ds485Ok = true;
  }
  else if (command == ledd_ds485_err) {
    m_ds485Ok = false;
  }
  else if (command == ledd_info2_ok) {
    m_info2Ok = true;
  }
  else if (command == ledd_info2_err) {
    m_info2Ok = false;
  }
  else if (command == ledd_info3_ok) {
    m_info3Ok = true;
  }
  else if (command == ledd_info3_err) {
    m_info3Ok = false;
  }
 
  // network substates
  else if (command == ledd_net_ok) {
    m_netOk = true;
  }
  else if (command == ledd_net_err) {
    m_netOk = false;
  }
  else if (command == ledd_wifi_ok) {
    m_wifiOk = true;
  }
  else if (command == ledd_wifi_err) {
    m_wifiOk = false;
  }
  else if (command == ledd_internet_ok) {
    m_internetOk = true;
  }
  else if (command == ledd_internet_err) {
    m_internetOk = false;
  }
  
  // failure substates
  else if (command == ledd_hw_ok) {
    m_hardwareErr = false;
  }
  else if (command == ledd_hw_err) {
    m_hardwareErr = true;
  }
  else if (command == ledd_sw_ok) {
    m_softwareErr = false;
  }
  else if (command == ledd_sw_err) {
    m_softwareErr = true;
  }
  else if (command == ledd_err2_ok) {
    m_err2Err = false;
  }
  else if (command == ledd_err2_err) {
    m_err2Err = true;
  }
  // ehem
  else {
    report(LOG_INFO, "Unknown or ignored command: %d - %s\n", command,
           ledd_action_to_string(command));
    return;
  }
 
  updateLeds();
}

void CommandHandler::handlerCallback(const ledd_actions command, void *arg) {
  try {
    if (arg == NULL) {
      report(LOG_ERR, "Ignoring command, missing object\n");
      return;
    }
    ((CommandHandler *)arg)->handle(command);
  } catch (...) {
    report(LOG_ERR, "Unhandled exception ignored\n");
  }
}
bool CommandHandler::updateLeds()
{
  segment* state;
  uint8_t ninfo;
  uint8_t nnetwork;
  uint8_t nerror;

  ninfo = 0;
  if (!m_ds485Ok) ninfo |= 1;
  if (!m_info2Ok) ninfo |= 2;
  if (!m_info3Ok) ninfo |= 4;

  nnetwork = 0;
  if (!m_netOk) nnetwork |= 1;
  if (!m_wifiOk) nnetwork |= 2;
  if (!m_internetOk) nnetwork |= 4;

  nerror = 0;
  if (m_hardwareErr) nerror |= 1;
  if (m_softwareErr) nerror |= 2;
  if (m_err2Err) nerror |= 4;
  
  switch (m_state) {
    case eSTATE_READY:
      state = code_ready;
      break;
    case eSTATE_FAILURE:
      state = code_failure;
      break;
   case eSTATE_STARTUP:
      state = code_startup;
      break;
   case eSTATE_UPDATE:
      state = code_system_update;
      ninfo = nnetwork = nerror = 0;
      break;
   case eSTATE_RAINBOW:
      state = code_rainbow;
      ninfo = nnetwork = nerror = 0;
      break;
  }
  
  Sequence* seq = new Sequence;
  if (seq == NULL) {
    report(LOG_ERR, "Failed to get memory for new sequence\n");
    return false;
  }

  uint32_t stateNo;
  
  stateNo = 0;
  while (state[stateNo].time) {
    LedGroup* group = new LedGroup();
    group->addLed(m_ledFactory->createLED("red", state[stateNo].red));
    group->addLed(m_ledFactory->createLED("green", state[stateNo].green));
    group->addLed(m_ledFactory->createLED("blue", state[stateNo].blue));
    seq->addLedGroup(group, state[stateNo].time);
    ++stateNo;
  }

  if (ninfo) {
    for (int n=1; n <= ninfo; n++) {
      LedGroup* group1 = new LedGroup();
      group1->addLed(m_ledFactory->createLED("red", 1));
      group1->addLed(m_ledFactory->createLED("green", 1));
      group1->addLed(m_ledFactory->createLED("blue", 1));
      seq->addLedGroup(group1, 500);
      LedGroup* group2 = new LedGroup();
      group2->addLed(m_ledFactory->createLED("red", state[stateNo-1].red));
      group2->addLed(m_ledFactory->createLED("green", state[stateNo-1].green));
      group2->addLed(m_ledFactory->createLED("blue", state[stateNo-1].blue));
      seq->addLedGroup(group2, 500);
    }
  }

  if (nerror) {
	stateNo = 0;
	while (state[stateNo].time) {
	  LedGroup* group = new LedGroup();
	  group->addLed(m_ledFactory->createLED("red", state[stateNo].red));
	  group->addLed(m_ledFactory->createLED("green", state[stateNo].green));
	  group->addLed(m_ledFactory->createLED("blue", state[stateNo].blue));
	  seq->addLedGroup(group, state[stateNo].time);
	  ++stateNo;
	}
	for (int n=1; n <= nerror; n++) {
      LedGroup* group1 = new LedGroup();
      group1->addLed(m_ledFactory->createLED("red", 1));
      group1->addLed(m_ledFactory->createLED("green", 0));
      seq->addLedGroup(group1, 500);
      LedGroup* group2 = new LedGroup();
      group2->addLed(m_ledFactory->createLED("red", 0));
      group2->addLed(m_ledFactory->createLED("green", state[stateNo-1].green));
      seq->addLedGroup(group2, 500);
    }
  }

  if (nnetwork) {
	stateNo = 0;
	while (state[stateNo].time) {
	  LedGroup* group = new LedGroup();
	  group->addLed(m_ledFactory->createLED("red", state[stateNo].red));
	  group->addLed(m_ledFactory->createLED("green", state[stateNo].green));
	  group->addLed(m_ledFactory->createLED("blue", state[stateNo].blue));
	  seq->addLedGroup(group, state[stateNo].time);
	  ++stateNo;
	}
	for (int n=1; n <= nnetwork; n++) {
      LedGroup* group1 = new LedGroup();
      group1->addLed(m_ledFactory->createLED("red", 1));
      group1->addLed(m_ledFactory->createLED("green", 0));
      group1->addLed(m_ledFactory->createLED("blue", 1));
      seq->addLedGroup(group1, 500);
      LedGroup* group2 = new LedGroup();
      group2->addLed(m_ledFactory->createLED("red", state[stateNo-1].red));
      group2->addLed(m_ledFactory->createLED("green", state[stateNo-1].green));
      group2->addLed(m_ledFactory->createLED("blue", state[stateNo-1].blue));
      seq->addLedGroup(group2, 500);
    }
  }

  stopSequence();
  m_sequence = seq;

  if (pthread_create(&m_thread, NULL, &Sequence::start, (void *) m_sequence)) {
	report(LOG_ERR, "Failed to start sequence thread!\n");
  }
  report(LOG_DEBUG, "started thread %p and sequence @%p\n", m_thread, m_sequence);
  return true;
}

void CommandHandler::stopSequence()
{
  if (m_sequence) {
    m_sequence->stop();
    pthread_join(m_thread, NULL);
    delete m_sequence;
    report(LOG_DEBUG, "stopped thread %p and sequence @%p\n", m_thread, m_sequence);
    m_sequence = 0;
    m_thread = 0;
  }
}
