/*
 * dSS11 led handling daemon
 * Copyright (c) 2011,2012 aizo ag, Switzerland
 *
 * Author: Johannes Winkelmann <johannes.winkelmann@aizo.com>
 *         Michael Tross <michael.tross@aizo.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <errno.h>

#include "sequence.h"
#include "ledgroup.h"
#include "report.h"


Sequence::Sequence()
  : m_stopRequested(false)
{
  pthread_mutex_init(&m_stopMutex, NULL);
  pthread_cond_init(&m_stopCond, NULL);
}

Sequence::~Sequence()
{
  reset();
  pthread_mutex_destroy(&m_stopMutex);
  pthread_cond_destroy(&m_stopCond);
}


void Sequence::reset()
{
  std::vector<std::pair<LedGroup*, uint32_t> >::iterator it = m_states.begin();
  for (; it != m_states.end(); ++it) {
    delete it->first;
    it->first = 0;
  }
  m_states.clear();
}



void Sequence::addLedGroup(LedGroup* ledGroup, uint32_t duration)
{
  m_states.push_back(std::make_pair(ledGroup, duration));
}

void* Sequence::start(void* p)
{
  Sequence* s = (Sequence*) p;
  struct timeval tv;
  struct timespec ts;

  if (s->m_stopRequested) {
    report(LOG_DEBUG, "activate nothing, stop requested\n");
    return NULL;
  }

  // special case of a steady state
  if (s->m_states.size() == 1) {
    report(LOG_DEBUG, "activate single state seq @%p\n", s);
    s->m_states.begin()->first->activate();
    return NULL;
  }

  report(LOG_DEBUG, "activate %d states seq @%p\n", s->m_states.size(), s);
  while (!s->m_stopRequested) {
    std::vector<std::pair<LedGroup*, uint32_t> >::iterator it = s->m_states.begin();
    for (; it != s->m_states.end() && !s->m_stopRequested; ++it) {

      it->first->activate();

      gettimeofday(&tv, NULL);
      ts.tv_sec = tv.tv_sec + it->second / 1000;
      ts.tv_nsec = tv.tv_usec * 1000 + (it->second % 1000) * 1000000;
      if (ts.tv_nsec >= 1000000000) {
        ts.tv_sec ++;
        ts.tv_nsec -= 1000000000;
      }

      int rc = 0;
      pthread_mutex_lock(&s->m_stopMutex);
      while (!s->m_stopRequested) {
        rc = pthread_cond_timedwait(&s->m_stopCond, &s->m_stopMutex, &ts);
        if (rc == ETIMEDOUT) {
          break;
        }
        else if (rc < 0) {
          break;
        }
      }
      pthread_mutex_unlock(&s->m_stopMutex);
    }
  }

  return NULL;
}

void Sequence::stop()
{
  pthread_mutex_lock(&m_stopMutex);
  m_stopRequested = true;
  pthread_cond_signal(&m_stopCond);
  pthread_mutex_unlock(&m_stopMutex);
}
